/*
 * Copyright 2000 Draagon Software, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Draagon Software, Inc.
 * Use is subject to license terms.
 *
 */

package com.draagon.wii;

import com.draagon.wii.internal.*;

import java.util.List;

/**
 * The WIIConfiguration interface is implemented by the configuration class
 * responsible for loading configuration data.  This class should be used when
 * constructing a custom method of managing configuration settings.  It is
 * currently called mainly in the <code>com.draagon.wii.servlet</code> package.
 *
 * @author Doug Mealing
 */
public interface WIIConfiguration
{
  /**
   * This method retrieves a general property setting for Web ii
   *
   * @param name  The name of the property to retrieve
   * @return the property value requested or <b>null</b> if not found
   */
  public String getProperty( String name );

  /**
   * This method retrieves the list of default request header classes.
   *
   * @return the Vector of request header classes
   */
  public List<String> getRequestHeaderClasses();

  /**
   * This method retrieves the list of request header classes that are used
   * specifically by the site being referenced.
   *
   * @param site  The site to retrieve the request header classes for
   * @return the Vector of request parameter classes
   */
  public List<String> getRequestHeaderClasses( WIISite site );

  /**
   * This method retrieves the list of default response header classes.
   *
   * @return the Vector of response header classes
   */
  public List<String> getResponseHeaderClasses();
  

  /**
   * This method retrieves the list of response header classes that are used
   * specifically by the site being referenced.
   *
   * @param site  The site to retrieve the header parameter classes for
   * @return the Vector of resonse header classes
   */
  public List<String> getResponseHeaderClasses( WIISite site );

  /**
   * This method retrieves the class name of the default connector to use.
   *
   * @return the classname for the default connector class
   */
  public String getConnectorClass();

  /**
   * This method retrieves the name of the connector class used by the
   * site specified.
   *
   * @param site  The site to retrieve the connector class for
   * @return the name of the connector class
   */
  public String getConnectorClass( WIISite site );

  /**
   * This method retrieves the name of the default stream class to use given
   * the specified content type
   *
   * @param content The content type to process
   * @return the name of the stream class
   */
  public String getStreamClass( String content );

  /**
   * This method retrieves the name of the stream class to use given the
   * specified content type and site
   *
   * @param site  The site to retrieve the stream class for
   * @param content The content type to process
   * @return the name of the stream class
   */
  public String getStreamClass( WIISite site, String content );

  /**
   * This method retrieves the name of the default service class requested
   * by name.
   *
   * @param service The name of the service to retrieve the class for
   * @return the name of the service class
   */
  public String getServiceClass( String service );

  /**
   * This method retrieves the name of the service class requested by name and
   * for the specified site.
   *
   * @param site  The site to retrieve the service class for
   * @param service The name of the service to retrieve the class for
   * @return the name of the service class
   */
  public String getServiceClass( WIISite site, String service );

  /**
   * This method retrieves the name of the service class requested by name and
   * for the specified site.
   *
   * @param site  The site to retrieve the service class for
   * @param service The name of the service to retrieve the class for
   * @return the name of the service class
   */
  public String getSiteProperty( WIISite site, String prop );

  public String getInternalServiceClass( String name );
  public String getInternalServiceProperty( InternalService is, String prop );

  public String getServiceProperty( WIIService service, String prop );
  public String getServiceProperty( WIISite site, WIIService service, String prop );

  public String getServiceTemplateProperty( WIITemplate template, WIIService service, String prop );

  public String getTemplateProperty( WIITemplate template, String prop );
}
