package com.draagon.wii.headers;

import com.draagon.wii.*;
import com.draagon.wii.util.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Title:        Draagon Web Integrator
 * Description:  Draagon Web Integrator is a web application integration framework.
 * Copyright:    Copyright (c) Doug Mealing
 * Company:      Draagon Software, Inc.
 * @author Doug Mealing
 * @version 1.0
 */

public class CookieModifier implements WIIRequestHeaders, WIIResponseHeaders
{
  private static Log log = LogFactory.getLog( CookieModifier.class );

  public CookieModifier()
  {
  }

  /**
   * Retrieves the new headers based on the headers from the HTTP request.
   * It will also add additional Draagon Web Integrator headers as needed
   *
   * @param req The WIIRequest object representing the client request which can
   *              be used to retrieve the client's request headers
   * @param headers The headers to be sent to the external system
   */
  public void processRequestHeaders( WIIRequest req, List<NameValuePair> headers ) throws WIIException
  {
	Collection<NameValuePair> toRemove = new ArrayList<NameValuePair>();
	
    for( NameValuePair pair : headers )
    {
      String name = pair.getName();

      if ( name.compareToIgnoreCase( "cookie" ) == 0 )
      {
        String val = pair.getValue();

        val = revertCookies( getCode( req ), val );
        if ( val == null )
        {
          log.error( "(processRequestHeaders) Invalid cookie format: [" + val + "]" );
        }
        else
        {
          if ( log.isDebugEnabled() )
            log.debug( "(processRequestHeaders) Reverting cookies to: [" + val + "]" );

          if ( !val.equals( "" )) {
            pair.set( name, val );
          }
          else {
        	  toRemove.add( pair );
          }
        }
      }
    }
    
    headers.remove( toRemove );
  }

  private String getCode( WIIRequest req )
  {
    String tmp = req.getUser().getId() + req.getSite().getName();
    int hash = tmp.hashCode();
    if ( hash < 0 ) hash = 0 - hash;
    return "" + hash;
  }

  private String revertCookies( String code, String val )
  {
    String tmp = null;
    String pre = code + "_";
    int pren = pre.length();

    if ( val == null || val.length() < 2 ) return null;

    int i = 0;
    while( true )
    {
      i = val.indexOf( ';' );
      String parse = val;

      if ( i == 0 ) return null;
      if ( i > 0 )
      {
        parse = val.substring( 0, i );
        val = val.substring( i + 1 );
        if ( val.length() < 2 ) val = null;
      }
      else
      {
        parse = val;
        val = null;
      }

      int j = parse.indexOf( '=' );
      if ( j <= 0 ) return null;

      String name = parse.substring( 0, j ).trim();
      String value = parse.substring( j + 1 );

      // Only pass back cookies that start with the prefix
      if ( name.startsWith( pre ))
      {
        name = name.substring( pren );

        if ( tmp == null )
          tmp = name + "=" + value;
        else
          tmp = tmp + "; " + name + "=" + value;
      }

      if ( val == null ) break;
    }

    if ( tmp == null ) return "";
    else return tmp;
  }

  private String convertCookie( String code, String val )
  {
    String end = null;

    if ( val == null || val.length() < 2 ) return null;

    int i = val.indexOf( ';' );
    if ( i == 0 ) return null;
    if ( i > 0 )
    {
      end = val.substring( i + 1 );
      val = val.substring( 0, i );
    }

    int j = val.indexOf( '=' );
    if ( j <= 0 ) return null;

    String name = val.substring( 0, j );
    String value = val.substring( j + 1 );

    name = code + "_" + name;

    if ( end != null )
      return name + "=" + value + ";" + end;
    else
      return name + "=" + value;
  }

  public void processResponseHeaders( WIIRequest req, WIIResponse res, List<NameValuePair> headers )
    throws WIIException
  {
    for( NameValuePair pair : headers )
    {
      String name = pair.getName();

      if ( name.compareToIgnoreCase( "set-cookie" ) == 0 ||
          name.compareToIgnoreCase( "set-cookie2" ) == 0 )
      {
        String val = pair.getValue();

        val = convertCookie( getCode( req ), val );
        if ( val == null )
        {
          log.error( "(processResponseHeaders) Invalid cookie format: [" + val + "]" );
        }
        else
        {
          if ( log.isDebugEnabled() )
            log.debug( "(processResponseHeaders) Converting cookie to: [" + val + "]" );

          pair.set( name, val );
        }
      }
    }
  }

  public static void main( String [] args )
  {
    CookieModifier c = new CookieModifier();
    System.out.println( "convert: " + c.convertCookie( "XXX", "doug=cool; timestamp=100; blah" ));
    System.out.println( "convert: " + c.convertCookie( "XXX", "doug=bad" ));
    System.out.println( "revert: " + c.revertCookies( "XXX", "session=A45kooRQ9Ym0nbH0; SESSIONID=To1012mC9061218840963297At" ));
    System.out.println( "revert: " + c.revertCookies( "XXX", "XXX_doug=cool; XXX_SESSIONID=To1012mC9061218840963297At" ));
    System.out.println( "revert: " + c.revertCookies( "XXX", "XXX_doug=bad" ));
  }
}
