/*
 * Copyright 2000 Draagon Software, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Draagon Software, Inc.
 * Use is subject to license terms.
 *
 */

package com.draagon.wii.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.net.*;
import java.io.*;

import javax.net.ssl.*;

public class WWWConnection extends InputStream
{
  private URL mURL = null;
  private Socket mSocket = null;
  private PushbackInputStream mIs = null;
  private OutputStream mOs = null;
  private HashMap<String,String> mRequestParams = new HashMap<String,String>();
  private List<NameValuePair> mResponseHeaders = new ArrayList<NameValuePair>();
  private String mRequestType = "GET";
  private boolean mInputCalled = false;
  private boolean mOutputCalled = false;
  @SuppressWarnings("unused")
  private String mHttpVersion = "";
  private int mResponseCode = 200;
  private String mResponseMessage = "";

  private String mServer = null;
  private int mPort = 0;

  // The If-Modified-Since: header could be called, so don't do this
  // protected String mContentType = "text/plain";
  private String mContentType = null;
  private boolean mIsChunked = false;

  private boolean mGotHeader = false;

  private int mContentLength = -1;
  public int mBytesRead = 0;
//  private int mByte = -1;

  private boolean mDebug = false;

  private boolean mUseSSL = false;

  public WWWConnection( URL url )
  {
    mURL = url;
  }

  public void open()
    throws IOException
  {
    if ( mURL.getProtocol().compareToIgnoreCase( "https" ) == 0 )
      mUseSSL = true;
    else if ( mURL.getProtocol().compareToIgnoreCase( "http" ) != 0 )
      throw new IOException( "Unsupported Protocol: " + mURL.getProtocol() );

    //////////////////////////////////////////////////
    // Create the default request parameters
    mRequestParams.put( "User-Agent", "OS/Web 0.6" );

    String host = mURL.getHost();
    if ( mURL.getPort() == -1 )
    {
      if ( mUseSSL ) host += ":443";
      else host += ":80";
    }
    else
      host += ":" + mURL.getPort();

    mRequestParams.put( "Host", host );
    mRequestParams.put( "Accept", "*/*" );
    mRequestParams.put( "Connection", "close" );
    //////////////////////////////////////////////////

    mServer = mURL.getHost();
    mPort = mURL.getPort();
    if ( mPort == -1 )
    {
      if ( mUseSSL ) mPort = 443;
      else mPort = 80;
    }
  }

  public void setRequestProperty( String name, String value )
  {
    if ( mDebug ) System.out.println( "setRequesProperty( " + name + ", " + value + ")" );

    if ( name == null ) mRequestType = value;
    // if ( name.compareToIgnoreCase( "connection" ) == 0 ) return;

    // Remove any duplicate items
    for( String key : mRequestParams.keySet() )
    {
      if ( key.compareToIgnoreCase( name ) == 0 )
      {
        mRequestParams.remove( key );
        break;
      }
    }

    mRequestParams.put( name, value );
  }

  public OutputStream getOutputStream()
    throws IOException
  {
    mOutputCalled = true;
    mRequestType = "POST";

    if ( mInputCalled )
      throw new IOException( "Already connected to server!" );

    return doConnection();
  }

  public InputStream getInputStream()
    throws IOException
  {
    mInputCalled = true;

    if ( !mOutputCalled ) doConnection();

    handleResponse();

    // if ( mContentLength < 0 ) throw new IOException( "No content-length defined!" );

    return this;
  }

  protected OutputStream doConnection()
    throws IOException
  {
    if ( mUseSSL )
    {
      SSLSocketFactory sslFact = (SSLSocketFactory) SSLSocketFactory.getDefault();
      mSocket = (SSLSocket) sslFact.createSocket( mServer, mPort );
    }
    else
      mSocket = new Socket( mServer, mPort );

    mOs = mSocket.getOutputStream();

    String line = mRequestType + " " + mURL.getFile() + " HTTP/1.1\r\n";

    String item = mURL.getFile();
    int i = item.lastIndexOf( '/' );
    if ( i >= 0 ) item = item.substring( i + 1 );

    if ( mDebug ) System.out.print( item + " >>> " + line );
    mOs.write( line.getBytes() );

    for ( String key : mRequestParams.keySet() )
    {
      String val = (String) mRequestParams.get( key );

      line = key + ": " + val + "\r\n";
      if ( mDebug ) System.out.print( item + " >>> " + line );
      mOs.write( line.getBytes() );
    }

    line = "\r\n";
    if ( mDebug ) System.out.print( ">>> " + line );
    mOs.write( line.getBytes() );
    mOs.flush();

    return mOs;
  }

  protected InputStream handleResponse()
    throws IOException
  {
    mIs = new PushbackInputStream( mSocket.getInputStream(), 512 );
    if ( mIs == null ) return null;

    while( true )
    {
      String line = readLine();
      if ( line == null ) return null;

      if ( line.length() == 0 )
      {
        if ( mResponseCode >= 100 && mResponseCode < 200 )
        {
          // This header should be ignored, and the request should have
          //  it's state reset and continue looking for a response header

          mRequestType = "GET";
          mResponseCode = 200;
          mResponseHeaders = new ArrayList<NameValuePair>();

          mHttpVersion = "";
          mResponseMessage = "";
          mContentType = null;
          mIsChunked = false;

          mContentLength = -1;
          mBytesRead = 0;

          mGotHeader = false;
        }
        else if ( mGotHeader )
        {
          if ( mIsChunked )
            mContentLength = getChunkedLength();

          return mIs;
        }
      }
      else
        handleLine( line );

      // System.out.println( "line [" + line + "]" );
    }
  }

  private int getChunkedLength()
    throws IOException
  {
    String line = null;

    while( true )
    {
      line = readLine();
      if ( line == null || line.trim().length() == 0 )
      {
        // System.out.println( "chunked EMPTY (" + line + ")" );
        continue;
      }
      break;
    }

    // System.out.println( "chunked PRE TRIM (" + line + ")" );

    line = line.trim();

    // System.out.println( "chunked -> [" + line + "]" );

    int val = 0;

    try {
      val = Integer.parseInt( line, 16 );
    }
    catch( NumberFormatException e ) {
      throw new IOException( "Improper chunk size line: [" + line + "]" );
    }

    // System.out.println( "chunked [" + line + "][" + val + "]" );

    if ( val == 0 )
    {
      while( true )
      {
        line = readLine();
        if ( line == null || line.trim().length() == 0 ) break;

        // System.out.println( "chunk header [" + line + "]" );
      }
    }

    return val;
  }

  private String readLine()
    throws IOException
  {
    StringBuffer buf = new StringBuffer();

    int rc = mIs.read();

    //int xx = 0; // Used to test if only a \n\n comes through

    while ( rc >= 0 )
    {
      if ( rc == 0 )
      {
        // Do nothing
      }
      else if ( rc == '\r' )
      {
        if ( buf.length() == 0 )
        {
          int x = mIs.read();  // get that last \n character
          if ( x < 0 ) return "";
          if ( x != '\n' ) mIs.unread( x );

          // System.out.println( "{" + x + "}" );

          return "";
        }
        else
        {
          String line = buf.toString();

          int x = mIs.read();  // get that last \n character
          if ( x < 0 ) return line;
          if ( x != '\n' )  mIs.unread( x );

          // System.out.println( "(" + x + ")" );

          return line;
        }
      }
      else if ( rc == '\n' )
      {
        // System.out.println( "*BAD*" );

        if ( buf.length() > 0 )
        {
          String line = buf.toString();
          return line;
        }

        return "";
      }
      else
        buf.append( (char) rc );

      rc = mIs.read();
    }

    return "";
  }

  public void handleLine( String in )
    throws IOException
  {
    if ( mDebug ) System.out.println( "<<< " + in );

    if ( in.startsWith( "HTTP" ))
    {
      int i = in.indexOf( ' ' );
      if ( i < 0 ) throw new IOException( "Could not parse HTTP protocol line: [" + in + "]" );

      mHttpVersion = in.substring( 0, i );
      in = in.substring( i + 1 );

      String code = "";

      i = in.indexOf( ' ' );
      if ( i < 0 )
        code = in;
      else
      {
        code = in.substring( 0, i );
        mResponseMessage = in.substring( i + 1 );
      }

      try
      {
        mResponseCode = Integer.parseInt( code );
      }
      catch( NumberFormatException e )
      {
        throw new IOException( "Could not parse HTTP protocol line: [" + in + "]" );
      }

      addResponseKey( null, mResponseCode + " " + mResponseMessage );

      mGotHeader = true;
    }
    else
    {
      int i = in.indexOf( ':' );
      if ( i < 0 ) throw new IOException( "Could not parse HTTP protocol line: [" + in + "]" );
      String key = in.substring( 0, i );
      String val = in.substring( i + 1 );

      addResponseKey( key.trim(), val.trim() );
    }
  }

  protected void addResponseKey( String key, String val )
    throws IOException
  {
    if ( key != null && key.compareToIgnoreCase( "content-type" ) == 0 )
      mContentType = val;
    else if ( key != null && key.compareToIgnoreCase( "content-length" ) == 0 )
    {
      try {
        mContentLength = Integer.parseInt( val );
      }
      catch ( NumberFormatException e ) {
        throw new IOException( "Invalid content length: (" + val + ")" );
      }
    }

    if ( key != null && key.compareToIgnoreCase( "transfer-encoding" ) == 0
        && val != null && val.compareToIgnoreCase( "chunked" ) == 0 )
    {
      mIsChunked = true;
    }
    else
    {
      mResponseHeaders.add( new NameValuePair( key, val ));
    }
  }

  public void close()
    throws IOException
  {
    if ( mIs != null ) mIs.close();
    if ( mOs != null ) mOs.close();
    mSocket.close();
  }

  public void setDebug( boolean flag )
  {
    mDebug = flag;
  }

  public List<NameValuePair> getHeaders()
  {
    return mResponseHeaders;
  }

  public String getContentType()
  {
    return mContentType;
  }

  public void setDoInput( boolean state )
  {
  }

  public void setDoOutput( boolean state )
  {
  }

  ////////////////////////////////////////////////////////////////////
  // Implemented to close InputStream when content length is reached!

  private int fill()
    throws IOException
  {
    if ( mContentLength > 0 && mBytesRead == mContentLength )
    {
      if (( mContentLength - mBytesRead ) == 0 && mIsChunked )
      {
        mContentLength = getChunkedLength();
        mBytesRead = 0;
      }
      else
        mContentLength = 0;
    }

    return mContentLength;
  }

  public int available()
    throws IOException
  {
    try
    {
      if ( fill() == 0 ) return -1;

      int x = mIs.available();
      if ( x != 0 )
      {
        if ( mContentLength > 0 && x > ( mContentLength - mBytesRead ))
          x = ( mContentLength - mBytesRead );

        return x;
      }

      int b = 0;
      if (( b = mIs.read() ) == -1 ) return -1;

      mIs.unread( b );
      return 1;
    }
    catch( java.net.SocketException e )
    {
      return -1;
    }
  }

  public int read()
    throws IOException
  {
    try
    {
      if ( fill() == 0 ) return -1;

      int x = mIs.read();
      if ( x >= 0 ) mBytesRead++;

      return x;
    }
    catch( java.net.SocketException e )
    {
      return -1;
    }
  }

  public int read( byte [] b )
    throws IOException
  {
    try
    {
      if ( fill() == 0 ) return -1;

      int x = 0;

      if ( mContentLength > 0 && b.length > ( mContentLength - mBytesRead ))
        x = mIs.read( b, 0, ( mContentLength - mBytesRead ));
      else
        x = mIs.read( b );

      if ( x > 0 ) mBytesRead += x;

      return x;
    }
    catch( java.net.SocketException e )
    {
      return -1;
    }
  }

  public int read( byte[] b, int i, int j )
    throws IOException
  {
    try
    {
      if ( fill() == 0 ) return -1;

      if ( mContentLength > 0 && j > ( mContentLength - mBytesRead ))
        j = mContentLength - mBytesRead;

      int x = mIs.read( b, i, j );
      if ( x >= 0 )
        mBytesRead += x;

      return x;
    }
    catch( java.net.SocketException e )
    {
      return -1;
    }
  }

  public void mark( int x ) {}
  public boolean markSupported() { return false; }
  public void reset() throws IOException {}
  public long skip( long x )  throws IOException { return x; }

  public int getStatusCode()
  {
    return mResponseCode;
  }
}
